from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
from tensorflow.python.tools import inspect_checkpoint as chkp
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import nn_ops
from tensorflow.python.ops import init_ops
from tensorflow.python.ops import variable_scope as vs
from tensorflow.python.framework import constant_op
from tensorflow.python.framework import dtypes

from keras import backend as K
from keras.models import load_model
from keras.preprocessing import image
from keras.optimizers import Adam
# from scipy.misc import imread
import numpy as np
from matplotlib import pyplot as plt

from keras.preprocessing import image as image_utils
from keras_ssd512 import ssd_512
from keras_ssd_loss import SSDLoss
from keras_layer_AnchorBoxes import AnchorBoxes
from keras_layer_L2Normalization import L2Normalization
from ssd_box_encode_decode_utils import decode_y, decode_y2
from ssd_batch_generator import BatchGenerator

import cv2, sys, os
import matplotlib.pyplot as plt
import imageio
import json, collections

from os import listdir
from os.path import isfile, join, isdir
from random import shuffle, choice
from PIL import Image

def warn(*args, **kwargs):
	pass
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

anchors = [(10,13), (16,30), (33,23), (30,61), (62,45), (59,119), (116,90), (156,198), (373,326)]

def non_max_suppression_fast(boxes, overlapThresh):
	# if there are no boxes, return an empty list
	if len(boxes) == 0:
		return []
	# if the bounding boxes integers, convert them to floats --
	# this is important since we'll be doing a bunch of divisions
	if boxes.dtype.kind == "i":
		boxes = boxes.astype("float")
	# initialize the list of picked indexes 
	pick = []
	# grab the coordinates of the bounding boxes
	x1 = boxes[:,0]
	y1 = boxes[:,1]
	x2 = boxes[:,2]
	y2 = boxes[:,3]
	# compute the area of the bounding boxes and sort the bounding
	# boxes by the bottom-right y-coordinate of the bounding box
	area = (x2 - x1 + 1) * (y2 - y1 + 1)
	idxs = np.argsort(y2)
	# keep looping while some indexes still remain in the indexes
	# list
	while len(idxs) > 0:
		# grab the last index in the indexes list and add the
		# index value to the list of picked indexes
		last = len(idxs) - 1
		i = idxs[last]
		pick.append(i)
		# find the largest (x, y) coordinates for the start of
		# the bounding box and the smallest (x, y) coordinates
		# for the end of the bounding box
		xx1 = np.maximum(x1[i], x1[idxs[:last]])
		yy1 = np.maximum(y1[i], y1[idxs[:last]])
		xx2 = np.minimum(x2[i], x2[idxs[:last]])
		yy2 = np.minimum(y2[i], y2[idxs[:last]])
		# compute the width and height of the bounding box
		w = np.maximum(0, xx2 - xx1 + 1)
		h = np.maximum(0, yy2 - yy1 + 1)
		# compute the ratio of overlap
		overlap = (w * h) / area[idxs[:last]]
		# delete all indexes from the index list that have
		idxs = np.delete(idxs, np.concatenate(([last],
			np.where(overlap > overlapThresh)[0])))
	# return only the bounding boxes that were picked using the
	# integer data type
	return boxes[pick].astype("int")

def loadData(jsonData, inPath):
	batchPaths = []
	for vidPath1 in jsonData["train_path"]:
		vidPath = join(inPath,vidPath1)
		framePath = os.path.join(os.path.dirname(vidPath), "frames")
		batchPaths = batchPaths + [framePath]
	return batchPaths

def loadMiniBatch(vidFilePath):
	vidName = vidFilePath.split('/')[-3]
	frameList = sorted([join(vidFilePath, f) for f in listdir(vidFilePath) if isfile(join(vidFilePath, f)) and f.endswith('.jpg')])
	frameList = sorted(frameList, key=lambda x: int(x.split('/')[-1].split('.')[0]))
	its = [iter(frameList), iter(frameList[1:])]
	segments = zip(*its)
	minibatch = []
	for segment in segments:
		im = []
		numFrames = 0
		for j, imFile in enumerate(segment):
			img = Image.open(imFile)
			img = img.resize((img_width, img_height), Image.ANTIALIAS)
			img = np.array(img)
			img = img[:, :, ::-1].copy() 
			# img = cv2.GaussianBlur(img,(5,5),0)
			im.append(img)
			numFrames += 1
		minibatch.append(np.stack(im))
	return vidFilePath, minibatch

# Set the image size.
img_height = 512
img_width = 512
num_channels = 3
newWidth = img_width
newHeight = img_height

slim = tf.contrib.slim
n_hidden1 = 512
n_hidden2 = 512
learnError = 0
n_epochs = 1
batch_size = 2

feature_size = 512
attention_shape = 32

seq_length = batch_size - 1

statefull_lstm_flag = False

lr_init = 1e-10

def broadcast(tensor, shape):
	return tensor + tf.zeros(shape, dtype=tensor.dtype)

def RNNCell(W, B, inputs, state):
	"""Most basic RNN: output = new_state = act(W * input + U * state + B)."""
	one = constant_op.constant(1, dtype=dtypes.int32)
	add = math_ops.add
	multiply = math_ops.multiply
	sigmoid = math_ops.sigmoid
	activation = math_ops.tanh

	gate_inputs = math_ops.matmul(array_ops.concat([inputs, state], 1), W)
	gate_inputs = nn_ops.bias_add(gate_inputs, B)
	output = sigmoid(gate_inputs)
	return output, output

def lstm_cell(W, b, forget_bias, inputs, state):
	one = constant_op.constant(1, dtype=dtypes.int32)
	add = math_ops.add
	multiply = math_ops.multiply
	sigmoid = math_ops.sigmoid
	activation = math_ops.sigmoid
	# activation = math_ops.tanh

	c, h = array_ops.split(value=state, num_or_size_splits=2, axis=one)

	gate_inputs = math_ops.matmul(array_ops.concat([inputs, h], 1), W)
	gate_inputs = nn_ops.bias_add(gate_inputs, b)
	# i = input_gate, j = new_input, f = forget_gate, o = output_gate
	i, j, f, o = array_ops.split(value=gate_inputs, num_or_size_splits=4, axis=one)

	forget_bias_tensor = constant_op.constant(forget_bias, dtype=f.dtype)

	new_c = add(multiply(c, sigmoid(add(f, forget_bias_tensor))), multiply(sigmoid(i), activation(j)))
	new_h = multiply(activation(new_c), sigmoid(o))
	new_state = array_ops.concat([new_c, new_h], 1)

	return new_h, new_state

class RNN_Predictor (tf.keras.Model):
	def __init__(self, batch_input_shape, input_shape, hidden_units):
	  
		super(RNN_Predictor, self).__init__()
	  
		# hidden state
		self.hidden_units = hidden_units
		
		self.hidden_state = tf.zeros((attention_shape*attention_shape, self.hidden_units))
		
		# BahdanauAttention layer to process input
		self.W1 = tf.keras.layers.Dense(hidden_units) 
		
		# BahdanauAttention layer to process hidden
		self.W2 = tf.keras.layers.Dense(hidden_units)

		# BahdanauAttention layer to generate context vector from embedded input and hidden state
		self.V = tf.keras.layers.Dense(1)

		# We also use a grid of 64 LSTMs to process the indiviual image blocks. The weights
		# of the LSTMs are the same. We implement this by passing the 64 inputs as
		# a "batch" into a single LSTM, so the batch size is not the default "BATCH_SIZE", which is 1
		# Note LSTM expects a 3D tensor (batch_size, seq_length, feature)
		self.LSTM1 = tf.keras.layers.LSTM(self.hidden_units,
										 batch_input_shape= (seq_length, batch_input_shape, input_shape),
										 time_major = True, #(timesteps, batch, ...)
										 # input_shape= (seq_length, input_shape),
										 return_sequences=True,
										 return_state=True, # return hidden state
										 stateful= statefull_lstm_flag)
		
		self.LSTM2 = tf.keras.layers.LSTM(self.hidden_units,
										 batch_input_shape= (seq_length, batch_input_shape, input_shape),
										 time_major = True, #(timesteps, batch, ...)
										 # input_shape= (seq_length, input_shape),
										 return_sequences=True,
										 return_state=True, # return hidden state
										 stateful= statefull_lstm_flag) 
		
		self.LSTM3 = tf.keras.layers.LSTM(self.hidden_units,
										 batch_input_shape= (seq_length, batch_input_shape, input_shape),
										 time_major = True, #(timesteps, batch, ...)
										 # input_shape= (seq_length, input_shape),
										 return_sequences=True,
										 return_state=True, # return hidden state
										 stateful= statefull_lstm_flag)	   
		
	def call(self, x): 
		# Dimension of x is [seq_length, attention_shape=64,  hidden_units] and is hidden_state is [attention_shape=64, hidden_units]

		# print ('In RNN Decoder x=', x.shape, 'hidden=', hidden_state.shape)

		# Bahadanou attention -- note we have hooks for temporal attention too.
				
		if (len(self.hidden_state.shape) == 2):
		  hidden_with_time_axis = tf.expand_dims(self.hidden_state, 0)
		else :
		  print ('Error: hidden_state should be 2 dimensional. It is:', hidden_state.shape)
		  hidden_with_time_axis = self.hidden_state
		
		# score shape == (seq_length, attention_shape=64, hidden_units)
		score = tf.nn.tanh(self.W1(x) + self.W2(hidden_with_time_axis))
		
		# attention_weights shape == (seq_length, attention_shape=64, 1)
		# you get 1 at the last axis because you are applying score to self.V
		attention_weights = tf.nn.softmax(self.V(score), axis=1)
		
		# context_vector shape == (seq_length, attention_shape=64, hidden_units)
		context_vector = tf.multiply(x, attention_weights)
		
		x_hat = context_vector  # (seq_length, attention_shape=64, embedding_dim)

		# LSTM expects a 3D tensor (seq_length, batch_size,  feature) since time_major == True


		encoder_output1, _ , _ = self.LSTM1 (x_hat)
		encoder_output2, _ , _ = self.LSTM2 (encoder_output1)
		output, next_hidden_state, cell_state = self.LSTM3 (encoder_output2)


		self.hidden_state = next_hidden_state

		#print ('x =', x_hat.shape, 'output=', output.shape, 'h state=', next_hidden_state.shape, 'cell state=', cell_state.shape)			   
			   
		return output, self.hidden_state, attention_weights

 
	def reset_hidden_state(self):
		self.hidden_state = tf.zeros((attention_shape*attention_shape, self.hidden_units))
		# LSTM expects a 3D tensor (batch_size, seq_length, feature)
		# we use batch_size = 64 (attention_shape) -- the 64 blocks of inceptionV3 features
		# seq_length, and feature = 2048 inceptionV3 features
		# hidden state dimension is (batch_size=attention_shape, features)

learnError = 0
@tf.function
def loss_function(target, pred, input_seq):  #Dimesions: [seq_length, 64, 2048]
  
	pred_loss = tf.square(tf.subtract(target, pred))
	
	# if we were to just do zero order hold as prediction, this is the loss or error
	frame_diff = tf.square(tf.subtract(target, input_seq)) * 0.01
	
	# prediction loss weighted by frame diffeence (zero-order hold difference), errors will be
	# weighted high for blocks with motion/change
	weighted_loss = tf.multiply(frame_diff, pred_loss) 
	
	sseLoss = tf.reduce_mean(weighted_loss)/(14*14*512)

	
	lossGrid = tf.nn.softmax(tf.reduce_mean(weighted_loss, 2), axis=1)
	
	return sseLoss, lossGrid

jsonData = json.load(open(sys.argv[1]))
vidPath = sys.argv[2]
modelPath = sys.argv[3]

if not os.path.exists(modelPath):
	os.makedirs(modelPath)

# TF Code
# inputs = tf.placeholder(tf.float32, (2, img_height, img_width, num_channels), name='inputs')
vgg16_Features = tf.placeholder(tf.float32, (2, attention_shape, attention_shape, feature_size), name='vgg16_Features')
bbPred = tf.placeholder(tf.float32, (2, 4), name='bb_inputs')
learning_rate = tf.placeholder(tf.float32, [])

pred_obj_feat = vs.get_variable("fpobj", shape=[1, feature_size])
pred_obj_geom = vs.get_variable("gpobj", shape=[1, 1, 4])

# Setup Feat LSTM
init_state1 = vs.get_variable("s1", shape=[2*n_hidden1])
W_lstm1 = vs.get_variable("W1", shape=[feature_size+n_hidden1, 4*n_hidden1])
b_lstm1 = vs.get_variable("b1", shape=[4*n_hidden1], initializer=init_ops.zeros_initializer(dtype=tf.float32))

W_fc1 = tf.Variable(tf.truncated_normal([n_hidden1, feature_size], stddev=0.1))
b_fc1 = tf.Variable(tf.constant(0.1, shape=[feature_size]))

curr_state1 = broadcast(init_state1, [1, 2*n_hidden1])

# Setup Geom LSTM
init_state2 = vs.get_variable("s2", shape=[2*n_hidden1])
W_lstm2 = vs.get_variable("W2", shape=[feature_size+n_hidden1, 4*n_hidden1])
b_lstm2 = vs.get_variable("b2", shape=[4*n_hidden1], initializer=init_ops.zeros_initializer(dtype=tf.float32))

W_fc2 = tf.Variable(tf.truncated_normal([n_hidden1, 4], stddev=0.1))
b_fc2 = tf.Variable(tf.constant(0.1, shape=[4]))

curr_state2 = broadcast(init_state1, [1, 2*n_hidden2])

predictModel = RNN_Predictor (batch_input_shape=attention_shape*attention_shape, input_shape=feature_size, hidden_units=n_hidden1)

input_sequence = tf.reshape(vgg16_Features[0,...], (-1, attention_shape*attention_shape, feature_size))
predictions, hidden, attention_weights = predictModel(input_sequence)

loss, loss_grid = loss_function(tf.reshape(vgg16_Features[1,...], (-1, attention_shape*attention_shape, feature_size)), predictions, input_sequence)

p_feats = tf.reshape(tf.keras.layers.GlobalAveragePooling1D()(attention_weights * predictions), (1, 1, -1))

feats = tf.reshape(tf.keras.layers.GlobalAveragePooling1D()(tf.reshape(loss_grid, (1, -1, 1)) * tf.reshape(vgg16_Features[0,...], (-1, attention_shape*attention_shape, feature_size))), (1, 1, -1))

lstm_inputs = array_ops.concat([tf.reshape(pred_obj_feat, [-1, feature_size])], 1)

h_1, curr_state1 = lstm_cell(W_lstm1, b_lstm1, 1.0, lstm_inputs, curr_state1)

h_2, curr_state2 = lstm_cell(W_lstm2, b_lstm2, 1.0, lstm_inputs, curr_state2)

pred_obj_feat = p_feats

query_value_attention_seq = tf.keras.layers.GlobalAveragePooling1D()(tf.reshape(tf.keras.layers.Attention()([feats, vgg16_Features[0,...]]), [1, -1, feature_size]))

fc1 = tf.matmul(h_1, W_fc1) + b_fc1
fc2 = (tf.matmul(h_2, W_fc2) + b_fc2) + bbPred[0,...]

diff = tf.reduce_mean(tf.square(tf.reshape(fc1, [-1, feature_size]) - query_value_attention_seq))
diff2 = tf.reduce_mean(tf.square(tf.reshape(fc2, [-1, 4]) - bbPred[1,...]))

# Optimization
total_loss = loss + diff + diff2*2
train_op = tf.train.GradientDescentOptimizer(learning_rate).minimize(total_loss)

batch = loadData(jsonData, vidPath)

#####################
### Training loop ###
#####################

init = tf.global_variables_initializer()
with tf.Session() as sess:
	sess.run(init)
	K.set_session(sess)

	model = ssd_512(image_size=(img_height, img_width, 3),
								n_classes=20,
								l2_regularization=0.0005,
								scales=[0.04, 0.1, 0.26, 0.42, 0.58, 0.74, 0.9, 1.06],
								aspect_ratios_per_layer=[[1.0, 2.0, 0.5],
																				 [1.0, 2.0, 0.5, 3.0, 1.0/3.0],
																				 [1.0, 2.0, 0.5, 3.0, 1.0/3.0],
																				 [1.0, 2.0, 0.5, 3.0, 1.0/3.0],
																				 [1.0, 2.0, 0.5, 3.0, 1.0/3.0],
																				 [1.0, 2.0, 0.5],
																				 [1.0, 2.0, 0.5]],
							 two_boxes_for_ar1=True,
							 steps=[8, 16, 32, 64, 128, 256, 512],
							 offsets=[0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5],
							 limit_boxes=False,
							 variances=[0.1, 0.1, 0.2, 0.2],
							 coords='centroids',
							 normalize_coords=True,
							 subtract_mean=[123, 117, 104],
							 swap_channels=True)

	# 2: Load the trained weights into the model.
	weights_path = './VGG_coco_SSD_512x512.h5'

	model.load_weights(weights_path, by_name=True)

	classes = ['background', 'aeroplane', 'bicycle', 'bird', 'boat', 'bottle', 'bus', 'car', 'cat', 'chair', 'cow', 'diningtable', 'dog', 'horse', 'motorbike', 'person', 'pottedplant', 'sheep', 'sofa', 'train', 'tvmonitor']

	saver = tf.train.Saver(max_to_keep=0)
	avgPredError = 1.0

	step = 0

	# LSTM
	new_state = np.random.uniform(-0.5,high=0.5,size=(1, attention_shape*attention_shape, 2*n_hidden1))
	for i in range(n_epochs):
		# Run 1 epoch
		loss1 = []
		shuffle(batch)

		for idx, miniBatchPath in enumerate(batch):

			new_state = np.random.uniform(-0.5,high=0.5,size=(1, attention_shape*attention_shape, 2*n_hidden1))

			lr = lr_init

			avgPredError = 0
			vidName, minibatches = loadMiniBatch(miniBatchPath)
			segCount = 0
			predError = collections.deque(maxlen=30)
			# print(miniBatchPath)
			for x_train in minibatches:
				img = x_train[0]
				segCount += 1

				vgg16Feat, bbPred_np = model.predict(x_train)
				
				bbPred_np = np.array(decode_y(bbPred_np, confidence_thresh=0.01, iou_threshold=0.45, top_k=200, input_coords='centroids', normalize_coords=True, img_height=img_height, img_width=img_width))[:,:,-4:]

				if segCount <= 1:
					bbPred_t_1 = np.reshape(bbPred_np[0,:2,...], (-1, 4))/512
				ret = sess.run([train_op, loss, fc2, attention_weights, total_loss, loss_grid], feed_dict = {vgg16_Features:vgg16Feat, bbPred:bbPred_t_1, learning_rate:lr})

				curr_attn = np.reshape(ret[-1], (-1, attention_shape, attention_shape))[0]
				curr_attn1 = np.reshape(ret[-3], (-1, attention_shape, attention_shape))[0]
				rects = []

				for box in bbPred_np[0]:
					xmin = box[-4]# * orig.shape[1] / img_width
					ymin = box[-3]# * orig.shape[0] / img_height
					xmax = box[-2]# * orig.shape[1] / img_width
					ymax = box[-1]# * orig.shape[0] / img_height
					rects.append([int(xmin), int(ymin), int(xmax), int(ymax)])
				
				
				if ret[1]/avgPredError > 1.5:
				 	lr = lr / 10.
				else:
				 	lr = lr_init
				predError.append(ret[1])
				avgPredError = np.mean(predError)
				loss1.append(ret[1])
				print('\rIdx:', idx, ret[1], ret[-2], ret[2], end="")

				a = np.reshape(ret[-1], (-1, attention_shape, attention_shape, 1))
				rects_final = []

				gridW, gridH = int(img_width/attention_shape), int(img_height/attention_shape)

				k = 5
				indices =  np.argpartition(a.flatten(), -2)[-k:]
				for _, y_pred, x_pred, _ in np.vstack(np.unravel_index(indices, a.shape)).T.tolist():
					predY = (y_pred)*gridH + int(0.5*gridH)
					predX = (x_pred)*gridW + int(0.5*gridW)

					rects_final += [(x1, y1, x2, y2) for x1, y1, x2, y2 in rects if x1 < predX < x2 and y1 < predY < y2]
				
				rects_final = list(set(rects_final))
				if rects_final:
					bbPred_t_1[0] = bbPred_t_1[1]
					bbPred_t_1[1] = np.reshape(np.array(rects_final[0]), (1, 4))/512

			print('\r Epoch = %d, Loss = %f'%(i, np.mean(loss1)), end="")
		path = modelPath + "_Epoch_" + str(i+1)	
		save_path = saver.save(sess, path)
